/*
 * Copyright (C) 2005-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file crypt.c openssl-based RSA encryption/decryption routines */

#include <openssl/pem.h>
#include "para.h"
#include "error.h"
#include "string.h"

#define LOAD_PUBLIC_KEY 0
#define LOAD_PRIVATE_KEY 1

static EVP_PKEY *load_key(const char *file, int private)
{
	BIO *key;
	EVP_PKEY *pkey = NULL;

	key = BIO_new(BIO_s_file());
	if (!key)
		return NULL;
	if (BIO_read_filename(key, file) > 0) {
		if (private == LOAD_PRIVATE_KEY)
			pkey = PEM_read_bio_PrivateKey(key, NULL, NULL, NULL);
		else
			pkey = PEM_read_bio_PUBKEY(key, NULL, NULL, NULL);
	}
	BIO_free(key);
	return pkey;
}

static int get_key(char *key_file, RSA **rsa, int private)
{
	EVP_PKEY *key = load_key(key_file, private);

	if (!key)
		return (private == LOAD_PRIVATE_KEY)? -E_PRIVATE_KEY : -E_PUBLIC_KEY;
	*rsa = EVP_PKEY_get1_RSA(key);
        EVP_PKEY_free(key);
        if (!*rsa)
		return -E_RSA;
	return RSA_size(*rsa);
}

int para_decrypt_buffer(char *key_file, unsigned char *outbuf, unsigned char *inbuf,
			int rsa_inlen)
{
	RSA *rsa;
	int ret = get_key(key_file, &rsa, LOAD_PRIVATE_KEY);

	if (ret < 0)
		return ret;
	ret = RSA_private_decrypt(rsa_inlen, inbuf, outbuf, rsa, RSA_PKCS1_PADDING);
	return (ret > 0)? ret : -E_DECRYPT;
}

int para_decrypt_challenge(char *key_file, long unsigned *challenge_nr,
		unsigned char *buf, int rsa_inlen)
{
	unsigned char *rsa_out = OPENSSL_malloc(128);
	int ret = para_decrypt_buffer(key_file, rsa_out, buf, rsa_inlen);

	if (ret < 0)
		return ret;
	ret = sscanf((char *)rsa_out, "%lu", challenge_nr) == 1?  1 : -E_CHALLENGE;
	OPENSSL_free(rsa_out);
	return ret;
}

int para_encrypt_buffer(char *key_file, unsigned char *inbuf,
		const unsigned len, unsigned char *outbuf)
{
	RSA *rsa;
	int ret = get_key(key_file, &rsa, LOAD_PUBLIC_KEY);

	if (ret < 0)
		return ret;
	ret = RSA_public_encrypt(len, inbuf, outbuf, rsa, RSA_PKCS1_PADDING);
	return ret < 0?  -E_ENCRYPT : ret;
}

int para_encrypt_challenge(char *key_file, long unsigned challenge_nr,
	unsigned char *outbuf)
{
	unsigned char *inbuf = (unsigned char*) make_message("%lu", challenge_nr);
	int ret = para_encrypt_buffer(key_file, inbuf, strlen((char *)inbuf), outbuf);
	free(inbuf);
	return ret;
}

